@@ -3,7 +3,7 @@
 #
 # This module is part of GitPython and is released under
 # the BSD License: http://www.opensource.org/licenses/bsd-license.php
-
+from __future__ import annotations
 from contextlib import contextmanager
 import io
 import logging
@@ -39,7 +39,7 @@ from .util import (
 
 # typing ---------------------------------------------------------------------------
 
-from typing import (Any, AnyStr, BinaryIO, Callable, Dict, IO, List, Mapping,
+from typing import (Any, AnyStr, BinaryIO, Callable, Dict, IO, Iterator, List, Mapping,
                     Sequence, TYPE_CHECKING, TextIO, Tuple, Union, cast, overload)
 
 from git.types import PathLike, Literal, TBD
@@ -68,7 +68,7 @@ __all__ = ('Git',)
 # Documentation
 ## @{
 
-def handle_process_output(process: subprocess.Popen,
+def handle_process_output(process: 'Git.AutoInterrupt' | Popen,
                           stdout_handler: Union[None,
                                                 Callable[[AnyStr], None],
                                                 Callable[[List[AnyStr]], None],
@@ -77,8 +77,9 @@ def handle_process_output(process: subprocess.Popen,
                                                 Callable[[AnyStr], None],
                                                 Callable[[List[AnyStr]], None]],
                           finalizer: Union[None,
-                                           Callable[[subprocess.Popen], None]] = None,
-                          decode_streams: bool = True) -> None:
+                                           Callable[[Union[subprocess.Popen, 'Git.AutoInterrupt']], None]] = None,
+                          decode_streams: bool = True,
+                          timeout: float = 10.0) -> None:
     """Registers for notifications to learn that process output is ready to read, and dispatches lines to
     the respective line handlers.
     This function returns once the finalizer returns
@@ -93,9 +94,10 @@ def handle_process_output(process: subprocess.Popen,
         their contents to handlers.
         Set it to False if `universal_newline == True` (then streams are in text-mode)
         or if decoding must happen later (i.e. for Diffs).
+    :param timeout: float, timeout to pass to t.join() in case it hangs. Default = 10.0 seconds
     """
     # Use 2 "pump" threads and wait for both to finish.
-    def pump_stream(cmdline: str, name: str, stream: Union[BinaryIO, TextIO], is_decode: bool,
+    def pump_stream(cmdline: List[str], name: str, stream: Union[BinaryIO, TextIO], is_decode: bool,
                     handler: Union[None, Callable[[Union[bytes, str]], None]]) -> None:
         try:
             for line in stream:
@@ -107,22 +109,32 @@ def handle_process_output(process: subprocess.Popen,
                     else:
                         handler(line)
         except Exception as ex:
-            log.error("Pumping %r of cmd(%s) failed due to: %r", name, remove_password_if_present(cmdline), ex)
-            raise CommandError(['<%s-pump>' % name] + remove_password_if_present(cmdline), ex) from ex
+            log.error(f"Pumping {name!r} of cmd({remove_password_if_present(cmdline)}) failed due to: {ex!r}")
+            raise CommandError([f'<{name}-pump>'] + remove_password_if_present(cmdline), ex) from ex
         finally:
             stream.close()
 
-    cmdline = getattr(process, 'args', '')  # PY3+ only
+    if hasattr(process, 'proc'):
+        process = cast('Git.AutoInterrupt', process)
+        cmdline: str | Tuple[str, ...] | List[str] = getattr(process.proc, 'args', '')
+        p_stdout = process.proc.stdout if process.proc else None
+        p_stderr = process.proc.stderr if process.proc else None
+    else:
+        process = cast(Popen, process)
+        cmdline = getattr(process, 'args', '')
+        p_stdout = process.stdout
+        p_stderr = process.stderr
+
     if not isinstance(cmdline, (tuple, list)):
         cmdline = cmdline.split()
 
-    pumps = []
-    if process.stdout:
-        pumps.append(('stdout', process.stdout, stdout_handler))
-    if process.stderr:
-        pumps.append(('stderr', process.stderr, stderr_handler))
+    pumps: List[Tuple[str, IO, Callable[..., None] | None]] = []
+    if p_stdout:
+        pumps.append(('stdout', p_stdout, stdout_handler))
+    if p_stderr:
+        pumps.append(('stderr', p_stderr, stderr_handler))
 
-    threads = []
+    threads: List[threading.Thread] = []
 
     for name, stream, handler in pumps:
         t = threading.Thread(target=pump_stream,
@@ -134,7 +146,9 @@ def handle_process_output(process: subprocess.Popen,
     ## FIXME: Why Join??  Will block if `stdin` needs feeding...
     #
     for t in threads:
-        t.join()
+        t.join(timeout=timeout)
+        if t.is_alive():
+            raise RuntimeError(f"Thread join() timed out in cmd.handle_process_output(). Timeout={timeout} seconds")
 
     if finalizer:
         return finalizer(process)
@@ -146,11 +160,11 @@ def dashify(string: str) -> str:
     return string.replace('_', '-')
 
 
-def slots_to_dict(self, exclude: Sequence[str] = ()) -> Dict[str, Any]:
+def slots_to_dict(self: object, exclude: Sequence[str] = ()) -> Dict[str, Any]:
     return {s: getattr(self, s) for s in self.__slots__ if s not in exclude}
 
 
-def dict_to_slots_and__excluded_are_none(self, d: Mapping[str, Any], excluded: Sequence[str] = ()) -> None:
+def dict_to_slots_and__excluded_are_none(self: object, d: Mapping[str, Any], excluded: Sequence[str] = ()) -> None:
     for k, v in d.items():
         setattr(self, k, v)
     for k in excluded:
@@ -165,7 +179,7 @@ CREATE_NO_WINDOW = 0x08000000
 ## CREATE_NEW_PROCESS_GROUP is needed to allow killing it afterwards,
 # see https://docs.python.org/3/library/subprocess.html#subprocess.Popen.send_signal
 PROC_CREATIONFLAGS = (CREATE_NO_WINDOW | subprocess.CREATE_NEW_PROCESS_GROUP  # type: ignore[attr-defined]
-                      if is_win else 0)
+                      if is_win else 0)                                       # mypy error if not windows
 
 
 class Git(LazyMixin):
@@ -192,7 +206,7 @@ class Git(LazyMixin):
     def __getstate__(self) -> Dict[str, Any]:
         return slots_to_dict(self, exclude=self._excluded_)
 
-    def __setstate__(self, d) -> None:
+    def __setstate__(self, d: Dict[str, Any]) -> None:
         dict_to_slots_and__excluded_are_none(self, d, excluded=self._excluded_)
 
     # CONFIGURATION
@@ -421,24 +435,27 @@ class Git(LazyMixin):
             return getattr(self.proc, attr)
 
         # TODO: Bad choice to mimic `proc.wait()` but with different args.
-        def wait(self, stderr: Union[None, bytes] = b'') -> int:
+        def wait(self, stderr: Union[None, str, bytes] = b'') -> int:
             """Wait for the process and return its status code.
 
             :param stderr: Previously read value of stderr, in case stderr is already closed.
             :warn: may deadlock if output or error pipes are used and not handled separately.
             :raise GitCommandError: if the return status is not 0"""
             if stderr is None:
-                stderr = b''
-            stderr = force_bytes(data=stderr, encoding='utf-8')
+                stderr_b = b''
+            stderr_b = force_bytes(data=stderr, encoding='utf-8')
 
             if self.proc is not None:
                 status = self.proc.wait()
 
-                def read_all_from_possibly_closed_stream(stream):
-                    try:
-                        return stderr + force_bytes(stream.read())
-                    except ValueError:
-                        return stderr or b''
+                def read_all_from_possibly_closed_stream(stream: Union[IO[bytes], None]) -> bytes:
+                    if stream:
+                        try:
+                            return stderr_b + force_bytes(stream.read())
+                        except ValueError:
+                            return stderr_b or b''
+                    else:
+                        return stderr_b or b''
 
                 if status != 0:
                     errstr = read_all_from_possibly_closed_stream(self.proc.stderr)
@@ -572,8 +589,8 @@ class Git(LazyMixin):
         self._environment: Dict[str, str] = {}
 
         # cached command slots
-        self.cat_file_header = None
-        self.cat_file_all = None
+        self.cat_file_header: Union[None, TBD] = None
+        self.cat_file_all: Union[None, TBD] = None
 
     def __getattr__(self, name: str) -> Any:
         """A convenience method as it allows to call the command as if it was
@@ -778,7 +795,12 @@ class Git(LazyMixin):
             log.info(' '.join(redacted_command))
 
         # Allow the user to have the command executed in their working dir.
-        cwd = self._working_dir or os.getcwd()
+        try:
+            cwd = self._working_dir or os.getcwd()  # type: Union[None, str]
+            if not os.access(str(cwd), os.X_OK):
+                cwd = None
+        except FileNotFoundError:
+            cwd = None
 
         # Start the process
         inline_env = env
@@ -907,7 +929,7 @@ class Git(LazyMixin):
         if self.GIT_PYTHON_TRACE == 'full':
             cmdstr = " ".join(redacted_command)
 
-            def as_text(stdout_value):
+            def as_text(stdout_value: Union[bytes, str]) -> str:
                 return not output_stream and safe_decode(stdout_value) or '<OUTPUT_STREAM>'
             # end
 
@@ -932,10 +954,10 @@ class Git(LazyMixin):
         else:
             return stdout_value
 
-    def environment(self):
+    def environment(self) -> Dict[str, str]:
         return self._environment
 
-    def update_environment(self, **kwargs):
+    def update_environment(self, **kwargs: Any) -> Dict[str, Union[str, None]]:
         """
         Set environment variables for future git invocations. Return all changed
         values in a format that can be passed back into this function to revert
@@ -962,7 +984,7 @@ class Git(LazyMixin):
         return old_env
 
     @contextmanager
-    def custom_environment(self, **kwargs):
+    def custom_environment(self, **kwargs: Any) -> Iterator[None]:
         """
         A context manager around the above ``update_environment`` method to restore the
         environment back to its previous state after operation.
@@ -1009,17 +1031,14 @@ class Git(LazyMixin):
 
     @classmethod
     def __unpack_args(cls, arg_list: Sequence[str]) -> List[str]:
-        if not isinstance(arg_list, (list, tuple)):
-            return [str(arg_list)]
 
         outlist = []
-        for arg in arg_list:
-            if isinstance(arg_list, (list, tuple)):
+        if isinstance(arg_list, (list, tuple)):
+            for arg in arg_list:
                 outlist.extend(cls.__unpack_args(arg))
-            # END recursion
-            else:
-                outlist.append(str(arg))
-        # END for each arg
+        else:
+            outlist.append(str(arg_list))
+
         return outlist
 
     def __call__(self, **kwargs: Any) -> 'Git':
@@ -1043,6 +1062,13 @@ class Git(LazyMixin):
                       ) -> str:
         ...  # if no args given, execute called with all defaults
 
+    @overload
+    def _call_process(self, method: str,
+                      istream: int,
+                      as_process: Literal[True],
+                      *args: Any, **kwargs: Any
+                      ) -> 'Git.AutoInterrupt': ...
+
     @overload
     def _call_process(self, method: str, *args: Any, **kwargs: Any
                       ) -> Union[str, bytes, Tuple[int, Union[str, bytes], str], 'Git.AutoInterrupt']:
@@ -1156,7 +1182,7 @@ class Git(LazyMixin):
         return refstr.encode(defenc)
 
     def _get_persistent_cmd(self, attr_name: str, cmd_name: str, *args: Any, **kwargs: Any
-                            ) -> Union['Git.AutoInterrupt', TBD]:
+                            ) -> 'Git.AutoInterrupt':
         cur_val = getattr(self, attr_name)
         if cur_val is not None:
             return cur_val
@@ -1166,12 +1192,16 @@ class Git(LazyMixin):
 
         cmd = self._call_process(cmd_name, *args, **options)
         setattr(self, attr_name, cmd)
+        cmd = cast('Git.AutoInterrupt', cmd)
         return cmd
 
-    def __get_object_header(self, cmd, ref: AnyStr) -> Tuple[str, str, int]:
-        cmd.stdin.write(self._prepare_ref(ref))
-        cmd.stdin.flush()
-        return self._parse_object_header(cmd.stdout.readline())
+    def __get_object_header(self, cmd: 'Git.AutoInterrupt', ref: AnyStr) -> Tuple[str, str, int]:
+        if cmd.stdin and cmd.stdout:
+            cmd.stdin.write(self._prepare_ref(ref))
+            cmd.stdin.flush()
+            return self._parse_object_header(cmd.stdout.readline())
+        else:
+            raise ValueError("cmd stdin was empty")
 
     def get_object_header(self, ref: str) -> Tuple[str, str, int]:
         """ Use this method to quickly examine the type and size of the object behind
@@ -1200,7 +1230,8 @@ class Git(LazyMixin):
         :note: This method is not threadsafe, you need one independent Command instance per thread to be safe !"""
         cmd = self._get_persistent_cmd("cat_file_all", "cat_file", batch=True)
         hexsha, typename, size = self.__get_object_header(cmd, ref)
-        return (hexsha, typename, size, self.CatFileContentStream(size, cmd.stdout))
+        cmd_stdout = cmd.stdout if cmd.stdout is not None else io.BytesIO()
+        return (hexsha, typename, size, self.CatFileContentStream(size, cmd_stdout))
 
     def clear_cache(self) -> 'Git':
         """Clear all kinds of internal caches to release resources.
